#!/usr/bin/perl

use strict;

if ($#ARGV < 1 || $#ARGV > 3) {
   print "Usage: $0 forward_results.out reverse_results.out [overlap_pct scorediff]\n";
   exit;
}

my ($fwd_file, $rev_file) = @ARGV; 

my $overlap_pct = 0.25; # what pct of reverse hit can be uncovered by a "covering" fwd hit
my $score_diff = 6;  # how many extra bits does a fwd hit have to offer before it's considered a true hit relative to the reverse sequence's false hit
if ($#ARGV > 2) {
  $overlap_pct = $ARGV[2];
  $score_diff = $ARGV[3];
}

my %f_aliends;
my %f_scores;
my %r_aliends;
my %r_qstarts;
my %r_qends;
my %r_scores;   
my %r_lines;   
   
$/ = "\n";

#read in hits on the forward strand, prepare them to be compared to reverse
open FH, "<$fwd_file";

my $sqlen = 60000;  # In this test, I know the length of all sequences is 60,000
my $line;
my ($target, $qstart, $qend, $alistart, $aliend, $score);
while ($line = <FH>) {

   my @vals = split (/\s+/, $line);
	
   ($target, $alistart, $aliend, $qstart, $qend, $score) = @vals[0,1,2,3,4,6];

   # for forward hits, flip around to where it would be on reverse
   my $tmp = $sqlen - $alistart + 1;
   $alistart   = $sqlen - $aliend + 1;
   $aliend = $tmp;

   $f_aliends{"$target,$alistart"}    = $aliend;
   $f_scores{"$target,$alistart"}     = $score;	
}

my @fwd_keys = sort by_name_start keys %f_aliends;

#merge overlapping hits on fwd. Do this so we don't fail to neutralize a reverse hit just 'cause it was at the boundary of two forward hits
my $j = 0;
my $i;

if ($#fwd_keys > 0 ) {
  for $i (1..$#fwd_keys) {
     my ($name, $start) = split (/,/, $fwd_keys[$i]);
     my $end            = $f_aliends{$fwd_keys[$i]} ;

     my ($prev_name, $prev_start) = split (/,/, $fwd_keys[$j]);
     my $prev_end                 = $f_aliends{$fwd_keys[$j]};

     if ( $name eq $prev_name &&  $start < $prev_end + 20) {  #if two hits are within 20 bases, merge them, keep the best score
        $f_aliends{$fwd_keys[$j]}  = $prev_end > $end ? $prev_end : $end; 
        $f_scores{$fwd_keys[$j]}  = $f_scores{$fwd_keys[$j]} > $f_scores{$fwd_keys[$i]} ? $f_scores{$fwd_keys[$j]} : $f_scores{$fwd_keys[$i]};
        #$merge_cnt++;
     } else {
        $j++;
        $fwd_keys[$j] = $fwd_keys[$i];
     }   
  }

  @fwd_keys = @fwd_keys[0..$j]; # compress list down
}




#read hits on reverse strand.  prepare for comparing to fwd
open FH, "<$rev_file";
my $rev_header = "";
my $rev_footer = "";
my $started = 0;
while (my $line = <FH>) {
   
   $started = 1;
   my @vals = split (/\s+/, $line);

   ($target, $alistart, $aliend, $qstart, $qend, $score) = @vals[0,1,2,3,4,6];	

	$r_aliends{"$target,$alistart"}   = $aliend;
	$r_qstarts{"$target,$alistart"}   = $qstart;	
	$r_qends{"$target,$alistart"}     = $qend;
	$r_scores{"$target,$alistart"}    = $score;
	$r_lines{"$target,$alistart"}     = $line;
}

my @rev_keys = sort by_name_start keys %r_aliends;


# walk lockstep along the two lists, deciding if a hit on reverse is a NP (also hit on fwd, but by longer hit)
# or a FP (either not hit on fwd, or hit but only by something not much longer than the rev hit) 
my $f = 0;
my $r = 0;
my $np_cnt = 0;
my @out_lines;

while ($f <= $#fwd_keys && $r <= $#rev_keys ) {
   
   my ($r_name,$r_start) = split(/,/,$rev_keys[$r]);
   my $r_end   = $r_aliends{$rev_keys[$r]};

   my ($f_name,$f_start) = split(/,/,$fwd_keys[$f]);
   my $f_end   = $f_aliends{$fwd_keys[$f]};

   #print "(1)$f_name v. $r_name,  $f_start..$f_end v  $r_start..$r_end\n";
   # skip to the next hit on fwd until we have one that either touches the reverse
   # hit or comes after it
   while ($f <= $#fwd_keys && ($f_name lt $r_name || $f_end < $r_start)) {
	  $f++;
          ($f_name,$f_start) = split(/,/,$fwd_keys[$f]);
	  $f_end   = $f_aliends{$fwd_keys[$f]};
   }
	
   my $buffer = $overlap_pct * ( $r_end - $r_start + 1); # allow fwd hits overlapping (1-overlap_pct) of the reverse hit to explain it as a NP:
   # overlap?
   if ($f_start <= $r_start + $buffer  && $f_end >= $r_end - $buffer && $f_scores{$fwd_keys[$f]} >= $r_scores{$rev_keys[$r]} + $score_diff) {
	  $np_cnt++;	
	  #print "np:   $f_start..$f_end   vs  $r_start..$r_end (buffer $buffer),  scores: $f_scores{$fwd_keys[$f]} v $r_scores{$rev_keys[$r]}\n";
   } else {
      # that's a false positive.  Lay it out on all the hmm positions:
	  #print "fp:   $f_start..$f_end   vs  $r_start..$r_end (buffer $buffer),  scores: $f_scores{$fwd_keys[$f]} v $r_scores{$rev_keys[$r]}\n";
      push @out_lines, $r_lines{$rev_keys[$r]};
   }
   $r++;
       
}

if ($#fwd_keys == -1) {
  foreach my $key ( keys %r_lines) { 
     push @out_lines, $r_lines{$key};
  }  
}


#print $rev_header;  #the CM lines kill my eyes
#@out_lines = sort by_score @out_lines;
print join ("", @out_lines);
#print $rev_footer;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  #
#   Subroutines             #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  #

sub by_name_start {
    my ($a_name, $a_start) = split (/,/,$a);
    my ($b_name, $b_start) = split (/,/,$b);
    
    if ( $a_name cmp $b_name ) { 
       return ($a_name cmp $b_name) ;
    } else {
       return ($a_start <=> $b_start);
    }
}

sub by_score {

   # extract the number from lines like:
   #gi|224384750|gb|CM000681.1|/10020001-10080000       1115  1241    2  127 +  109.6 2.5e-29
   
   my ($s_a) = ($a =~ / (\d+\.\d+) /);
   my ($s_b) = ($b =~ / (\d+\.\d+) /);	

   return ($s_a <=> $s_b);
}


sub do_cmd {
    my $cmd = shift;
#	print STDERR "$cmd\n";
	return `$cmd`;
}

